//
// Access.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

/**
 * Decodes an access record within a Java <tt>.class</tt> file.
 *
 * @version 1.1, 01 September 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John Leach</a>
**/

public final class Access {

 /** The access record size. */
 static int size = 2;

 /** The public access constant. */
 private static final int PUBLIC =	 0x0001;

 /** The private access constant. */
 private static final int PRIVATE =	 0x0002;

 /** The protected access constant. */
 private static final int PROTECTED =	 0x0004;

 /** The static access constant. */
 private static final int STATIC =	 0x0008;

 /** The final access constant. */
 private static final int FINAL =	 0x0010;

 /** The synchronized access constant. */
 private static final int SYNCHRONIZED = 0x0020;

 /** The volatile access constant. */
 private static final int VOLATILE =	 0x0040;

 /** The transient access constant. */
 private static final int TRANSIENT =	 0x0080;

 /** The native access constant. */
 private static final int NATIVE =	 0x0100;

 /** The interface access constant. */
 private static final int INTERFACE =	 0x0200;

 /** The abstract access constant. */
 private static final int ABSTRACT =	 0x0400;

 /** The access flag. */
 private int accessFlag;

 /** The access flag as a string. */
 private String accessStr;

 /**
  * Creates the access flag from a byte array.
  * @param buffer     the byte array
  * @param index      the offset in the byte array
  * @exception ClassException	    if the access flag cannot be decoded
 **/

 public Access(byte[] buffer, int index) throws ClassException {
  try {
   accessFlag = ValueDecoder.readU2(buffer, index);
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException("Access flag corrupt: " + index);
   }
  }

 /**
  * Checks if the access is public.
  * @return	   true if public, otherwise false
 **/

 public boolean isPublic() {
  return ((accessFlag & PUBLIC) != 0);
  }

 /**
  * Checks if the access is private.
  * @return	   true if private, otherwise false
 **/

 public boolean isPrivate() {
  return ((accessFlag & PRIVATE) != 0);
  }

 /**
  * Checks if the access is protected.
  * @return	   true if protected, otherwise false
 **/

 public boolean isProtected() {
  return ((accessFlag & PROTECTED) != 0);
  }

 /**
  * Checks if the access is static.
  * @return	   true if static, otherwise false
 **/

 public boolean isStatic() {
  return ((accessFlag & STATIC) != 0);
  }

 /**
  * Checks if the access is final.
  * @return	   true if final, otherwise false
 **/

 public boolean isFinal() {
  return ((accessFlag & FINAL) != 0);
  }

 /**
  * Checks if the access is synchronised.
  * @return	   true if synchronised, otherwise false
 **/

 public boolean isSynchronized() {
  return ((accessFlag & SYNCHRONIZED) != 0);
  }

 /**
  * Checks if the access is volatile.
  * @return	   true if volatile, otherwise false
 **/

 public boolean isVolatile() {
  return ((accessFlag & VOLATILE) != 0);
  }

 /**
  * Checks if the access is transient.
  * @return	   true if transient, otherwise false
 **/

 public boolean isTransient() {
  return ((accessFlag & TRANSIENT) != 0);
  }

 /**
  * Checks if the access is native.
  * @return	   true if native, otherwise false
 **/

 public boolean isNative() {
  return ((accessFlag & NATIVE) != 0);
  }

 /**
  * Checks for an interface or a class.
  * @return	   true if interface, false if class
 **/

 public boolean isInterface() {
  return ((accessFlag & INTERFACE) != 0);
  }

 /**
  * Checks if the access is abstract.
  * @return	   true if abstract, otherwise false
 **/

 public boolean isAbstract() {
  return ((accessFlag & ABSTRACT) != 0);
  }

 /**
  * Converts the access flag to a string.
 **/

 public String stringize() {
  if(accessStr == null) {
   accessStr = "";

   if(isPublic())	  accessStr = accessStr + "public ";
   else if(isPrivate())   accessStr = accessStr + "private ";
   else if(isProtected()) accessStr = accessStr + "protected ";

   if(isStatic())	  accessStr = accessStr + "static ";
   if(isFinal())	  accessStr = accessStr + "final ";
   if(isSynchronized())   accessStr = accessStr + "synchronized ";
   if(isVolatile())	  accessStr = accessStr + "volatile ";
   if(isTransient())	  accessStr = accessStr + "transient ";
   if(isNative())	  accessStr = accessStr + "native ";
   if(isAbstract())	  accessStr = accessStr + "abstract ";
   if(isInterface())	  accessStr = accessStr + "interface ";
   }

  return accessStr;
  }
 }
