//
// AttributeFactory.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

/**
 * A factory (object generator) for any attribute record in a Java
 * <tt>.class</tt> file.
 *
 * @version 1.0, 01 September 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John Leach</a>
 * @see Attribute
 * @see ConstantValueAttribute
 * @see CodeAttribute
 * @see ExceptionsAttribute
 * @see LineNumberTableAttribute
 * @see LocalVariableTableAttribute
 * @see SourceFileAttribute
 * @see UnknownAttribute
**/

public final class AttributeFactory {

 /** The buffer offset. */
 public int offset;

 /** The attribute size. */
 public int size;

 /** The constructed attribute record. */
 public Attribute record;

 /** The exception error message. */
 private static final String error = "Attribute record corrupt: ";

 /**
  * Creates a series of attribute records from a byte array.
  * @param buffer     the byte array
  * @param index      the offset in the input buffer
  * @exception ClassException	    if the record cannot be decoded
 **/

 public AttributeFactory(byte[] buffer, int index) throws ClassException {

  String name;
  int nameIndex;

  try {
   nameIndex = ValueDecoder.readU2(buffer, index);
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException(error + index);
   }

  try {
   if(Constant.constantPool[nameIndex].type == Constant.UTF8) name = Constant.constantPool[nameIndex].stringize();
   else throw new ClassException(error + index + " [wrong type]");
   }
  catch(NullPointerException e) {
   throw new ClassException(error + index + " [no entry]");
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException(error + index + " [no entry]");
   }

  index += 2;

  if(name.equals("ConstantValue"))	     record = new ConstantValueAttribute(buffer, index);
  else if(name.equals("Code"))		     record = new CodeAttribute(buffer, index);
  else if(name.equals("Exceptions"))	     record = new ExceptionsAttribute(buffer, index);
  else if(name.equals("LineNumberTable"))    record = new LineNumberTableAttribute(buffer, index);
  else if(name.equals("LocalVariableTable")) record = new LocalVariableTableAttribute(buffer, index);
  else if(name.equals("SourceFile"))	     record = new SourceFileAttribute(buffer, index);
  else					     record = new UnknownAttribute(buffer, index, name);

  // update the size and offset

  size = record.size + 2;
  offset = index + record.size;
  }
 }
