//
// ClassData.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

import java.io.*;
import java.util.Vector;

/**
 * A container for the textual information of a Java <tt>.class</tt> file.
 * <br>
 * The contents of the jclass.ClassData are set by the class jclass.Parser 
 * or a sub-class thereof.
 *
 * @version 1.1, 10 October 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John Leach</a>
 * @see Parser
**/

public class ClassData {

 /** The compiler version number. */
 protected String versionNumber;

 /** The .class filename. */
 protected String fileName;

 /** The class source filename. */
 protected String sourceFileName;

 /** The class name. */
 protected String className;

 /** The super-class name. */
 protected String superName;

 /** The classes used (imported) by the .class file. */
 protected Vector usedClasses;

 /** The full class specification. */
 protected String classSpec;

 /** The class fields. */
 protected Vector classFields;

 /** The class methods. */
 protected Vector classMethods;

 /** The field references. */
 protected Vector fieldRefs;

 /** The method references. */
 protected Vector methodRefs;

 /** The interface method references. */
 protected Vector interfaceMethodRefs;

 /**
  * Creates a ClassData object.
  * @see Parser
 **/

 public ClassData() {
  }

 /**
  * Sets the compiler version number, since this is a public method, and the
  * version number should not be changed after parsing, the compiler version
  * number can only be set once.
  * @param ver	   the version number
 **/

 public void setVersion(String ver) {
  if(versionNumber == null) versionNumber = ver;
  }

 /**
  * Gets the version number.
  * @return	the compiler version number, or "" if not set
 **/

 public String getVersion() {
  return (versionNumber == null) ? "" : versionNumber;
  }

 /**
  * Sets the soure filename, since this is a public method, and the source 
  * filename should not be changed after parsing, the source filename can 
  * only be set once.
  * @param name     the source file name
 **/

 public void setSourceFileName(String name) {
  if(sourceFileName == null) sourceFileName = name;
  }

 /**
  * Gets the source filename.
  * @return	the source filename, or "" if not set
 **/

 public String getSourceFileName() {
  return (sourceFileName == null) ? "" : sourceFileName;
  }

 /**
  * Sets the <tt>.class</tt> filename.
  * @param name     the .class file name
 **/

 public void setFileName(String name) {
  if(fileName == null) fileName = name;
  }

 /**
  * Gets the <tt>.class</tt> filename.
  * @return	the .class filename, or "" if not set
 **/

 public String getFileName() {
  return (fileName == null) ? "" : fileName;
  }

 /**
  * Sets the class name, since this is a public method, and the class name
  * should not be changed after parsing, the class name can only be set once.
  * @param name     the class name
 **/

 public void setClassName(String name) {
  if(className == null) className = name;
  }

 /**
  * Gets the class name.
  * @return	the class name, or "" if not set
 **/

 public String getClassName() {
  return (className == null) ? "" : className;
  }

 /**
  * Sets the super class name, since this is a public method, and the super
  * class name should not be changed after parsing, the super class name can
  * only be set once.
  * @param name     the super class name
 **/

 public void setSuperClassName(String name) {
  if(superName == null) superName = name;
  }

 /**
  * Gets the super class name.
  * @return	the super class name, or "" if not set
 **/

 public String getSuperClassName() {
  return (superName == null) ? "" : superName;
  }

 /**
  * Sets the used (imported) class names, since this is a public method, and
  * the used class names should not be changed after parsing, the used class
  * names can only be set once.
  * @param names     the used class names array
 **/

 public void setUsedClasses(String names[]) {
  if(usedClasses == null) usedClasses = orderNames(names);
  }

 /**
  * Checks if this class uses other classes.
  * @return	true if other classes are used, otherwise false
 **/

 public boolean usesClasses() {
  if(usedClasses == null || usedClasses.size() == 0) return false;
  return true;
  }

 /**
  * Gets the used classes name vector.
  * @return	the used classes vector, or null if not set
 **/

 public Vector getUsedClasses() { return usedClasses; }

 /**
  * Sets the class specification, since this is a public method, and the
  * class specifiation should not be changed after parsing, the class
  * specification can only be set once.
  * @param spec     the class specification
 **/

 public void setClassSpec(String spec) {
  if(classSpec == null) classSpec = spec;
  }

 /**
  * Gets the class specification.
  * @return	the class specification, or "" if not set
 **/

 public String getClassSpec() {
  return (classSpec == null) ? "" : classSpec;
  }

 /**
  * Sets the class field names, since this is a public method, and the class
  * field names should not be changed after parsing, the class field names can
  * only be set once.
  * @param names     the class field names array
 **/

 public void setClassFields(String names[]) {
  if(classFields == null) classFields = orderNames(names);
  }

 /**
  * Checks if this class has fields.
  * @return	true if this class has fields, otherwise false
 **/

 public boolean hasFields() {
  if(classFields == null || classFields.size() == 0) return false;
  return true;
  }

 /**
  * Gets the class field names vector.
  * @return	the class field names vector, or null if not set
 **/

 public Vector getClassFields() { return classFields; }

 /**
  * Sets the class method names, since this is a public method, and the class
  * method names should not be changed after parsing, the class method names
  * can only be set once.
  * <br>
  * Note: the setClassName() method should be called first, otherwise
  * constructors will not be ordered at the top of the list.
  * @param names     the class method names array
  * @see #setClassName
 **/

 public void setClassMethods(String names[]) {
  if(classMethods == null) classMethods = orderMethods(names);
  }

 /**
  * Checks if this class has methods.
  * @return	true if this class has methods, otherwise false
 **/

 public boolean hasMethods() {
  if(classMethods == null || classMethods.size() == 0) return false;
  return true;
  }

 /**
  * Gets the class method names vector.
  * @return	the class method names vector, or null if not set
 **/

 public Vector getClassMethods() { return classMethods; }

 /**
  * Sets the field references, since this is a public method, and the field
  * references should not be changed after parsing, the field references can
  * only be set once.
  * @param refs     the field references array
 **/

 public void setFieldRefs(String refs[]) {
  if(fieldRefs == null) fieldRefs = orderNames(refs);
  }

 /**
  * Checks if this class has field references.
  * @return	true if this class has field references, otherwise false
 **/

 public boolean hasFieldRefs() {
  if(fieldRefs == null || fieldRefs.size() == 0) return false;
  return true;
  }

 /**
  * Gets the field references vector.
  * @return	the field references vector, or null if not set
 **/

 public Vector getFieldRefs() { return fieldRefs; }

 /**
  * Sets the method references, since this is a public method, and the method
  * references should not be changed after parsing, the method references can
  * only be set once.
  * <br>
  * Note: the setClassName() method should be called first, otherwise
  * constructors will not be ordered at the top of the list
  * @param refs     the method references array
  * @see #setClassName
 **/

 public void setMethodRefs(String refs[]) {
  if(methodRefs == null) methodRefs = orderImportedMethods(refs);
  }

 /**
  * Checks if this class has method references.
  * @return	true if this class has method references, otherwise false
 **/

 public boolean hasMethodRefs() {
  if(methodRefs == null || methodRefs.size() == 0) return false;
  return true;
  }

 /**
  * Gets the method references vector.
  * @return	the method references vector, or null if not set
 **/

 public Vector getMethodRefs() { return methodRefs; }

 /**
  * Sets the interface method references, since this is a public method, and
  * the interface method references should not be changed after parsing, the
  * interface method references can only be set once.
  * @param refs     the interface method references array
 **/

 public void setInterfaceMethodRefs(String refs[]) {
  if(interfaceMethodRefs == null) interfaceMethodRefs = orderNames(refs);
  }

 /**
  * Check if this class has interface method references.
  * @return	true if this class has interface method references, otherwise
  *		false
 **/

 public boolean hasInterfaceMethodRefs() {
  if(interfaceMethodRefs == null || interfaceMethodRefs.size() == 0) return false;
  return true;
  }

 /**
  * Gets the interface method references vector.
  * @return	the interface method references vector, or null if not set
 **/

 public Vector getInterfaceMethodRefs() { return interfaceMethodRefs; }

 /**
  * Orders an array of names.
  * @param names    the string array of names
  * @return	    the ordered vector
 **/

 private Vector orderNames(String[] names) {
  Vector result = new Vector();
  if(names == null || names.length == 0) return result;

  Vector identifiers = new Vector();
  String name;
  int index;
  for(int nameIndex = 0; nameIndex < names.length; nameIndex++) {

   // skip null strings

   name = names[nameIndex];
   if(name == null) continue;

   index = addName(name, identifiers);
   result.insertElementAt(name, index);
   }

  return result;
  }

 /**
  * Orders an array of methods, constructors first.
  * <br>
  * Note: degenerates to orderNames() if the class name has not been set.
  * @param names    the string array of methods
  * @return	    the ordered vector
 **/

 private Vector orderMethods(String[] names) {
  if(className == null) return orderNames(names);

  Vector constructors = new Vector();
  if(names == null || names.length == 0) return constructors;

  Vector methods = new Vector();
  Vector cIdentifiers = new Vector();
  Vector mIdentifiers = new Vector();
  String name;
  int index;

  for(int nameIndex = 0; nameIndex < names.length; nameIndex++) {

   // skip null strings

   name = names[nameIndex];
   if(names[nameIndex] == null) continue;

   if(parseClassName(name).startsWith(className)) {
    index = addName(name, cIdentifiers);
    constructors.insertElementAt(name, index);
    }
   else {
    index = addName(name, mIdentifiers);
    methods.insertElementAt(name, index);
    }
   }

  for(int i = 0; i < methods.size(); i++) constructors.addElement(methods.elementAt(i));
  return constructors;
  }

 /**
  * Orders an array of methods, classes constructors first, then imported
  * constructors, then methods.
  * @param names    the string array of methods
  * @return	    the ordered vector
 **/

 private Vector orderImportedMethods(String[] names) {
  Vector constructors = new Vector();

  if(names == null || names.length == 0) return constructors;

  Vector importConstructors = new Vector();
  Vector methods = new Vector();
  Vector cIdentifiers = new Vector();
  Vector iIdentifiers = new Vector();
  Vector mIdentifiers = new Vector();
  String name;
  int index;

  for(int nameIndex = 0; nameIndex < names.length; nameIndex++) {

   // skip null strings

   name = names[nameIndex];
   if(name == null) continue;

   if(parseClassName(name).equals(className)) {
    index = addName(name, cIdentifiers);
    constructors.insertElementAt(name, index);
    }
   else if(isImportName(name)) {
    index = addName(name, iIdentifiers);
    importConstructors.insertElementAt(name, index);
    }
   else {
    index = addName(name, mIdentifiers);
    methods.insertElementAt(name, index);
    }
   }

  for(int i = 0; i < importConstructors.size(); i++) constructors.addElement(importConstructors.elementAt(i));
  for(int i = 0; i < methods.size(); i++) constructors.addElement(methods.elementAt(i));
  return constructors;
  }

 /**
  * Adds a string to a vector in alphabetical order.
  * @param name        the string to add
  * @param vect        the string vector
  * @return	the insertion index
 **/

 private int addName(String name, Vector vect) {
  int upper, lower, key, comparison;
  String parsedName = parseName(name);
  String compName;

  key = comparison = 0;
  if((upper = vect.size()) > 0) {
   lower = -1;
   while(true) {
    key = (upper + lower) >> 1;
    compName = (String)vect.elementAt(key);
    if((comparison = parsedName.compareTo(compName)) == 0) break;
    if(comparison < 0) upper = key;
    else lower = key;
    if(lower >= (upper - 1)) break;
    }
   }
  if(comparison > 0) ++key;
  vect.insertElementAt(parsedName, key);
  return key;
  }

 /**
  * Strips leading access values.
  * @param name     the name to strip
  * @return	    the stripped name
 **/

 private String parseName(String name) {
  int index = name.indexOf(' ');
  int lastIndex = 0;
  int closeIndex = name.indexOf(')');
  int equalIndex = name.indexOf(" =");

  if(closeIndex < 0) closeIndex = name.length();
  if(equalIndex < 0) equalIndex = name.length();

  int maxIndex = (closeIndex < equalIndex) ? closeIndex : equalIndex;

  while(index < maxIndex && index > 0) {
   lastIndex = index;
   index = name.indexOf(' ', index + 1);
   }

  return name.substring(lastIndex, maxIndex);
  }

 /**
  * Strips leading access values, and parameters.
  * @param name     the name to strip
  * @return	    the stripped name
 **/

 private String parseClassName(String name) {
  String parsedName = parseName(name);
  int index = parsedName.indexOf('(');

  if(index < 0) return parsedName;
  return parsedName.substring(0, index);
  }

 /**
  * Checks if the method name matches an import class name.
  * @param name     the name to check
  * @return	    true if match found, otherwise false
 **/

 private boolean isImportName(String name) {
  if(usedClasses == null || usedClasses.size() == 0) return false;

  String parsedName = parseClassName(name);

  String compName;
  int key = 0;
  int upper = usedClasses.size();
  int lower = -1;
  int comparison;
  while(true) {
   key = (upper + lower) >> 1;
   compName = (String)usedClasses.elementAt(key);
   if((comparison = parsedName.compareTo(compName)) == 0) return true;
   if(comparison < 0) upper = key;
   else lower = key;
   if(lower >= (upper - 1)) return false;
   }
  }
 }
