//
// CodeAttribute.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

/**
 * Decodes a Code attribute record in a Java <tt>.class</tt> file
 *
 * @version 1.0, 02 September 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John Leach</a>
 * @see ExceptionElement
 * @see Attribute
 * @see AttributeFactory
 * @see SourceFileAttribute
 * @see ConstantValueAttribute
 * @see ExceptionsAttribute
 * @see LineNumberTableAttribute
 * @see LocalVariableTableAttribute
 * @see UnknownAttribute
**/

public final class CodeAttribute extends Attribute {

 /** The exception error message. */
 private static final String error = "Code attribute record corrupt: ";

 /**
  * Creates a Code attribute record from a byte array.
  * @param buffer     the byte array
  * @param index      the offset in the input buffer
  * @exception ClassException	if the record cannot be decoded
 **/

 public CodeAttribute(byte[] buffer, int index) throws ClassException {

  try {

   // NOTE: the size is an unsigned 32 bit value in the specs.

   size = ValueDecoder.readU4(buffer, index);
   if(size <= 0) throw new ClassException(error + index + " [size <= 0]");
   index += 4;

   // used to check that the size is correct

   int orgIndex = index;

   int maxStack = ValueDecoder.readU2(buffer, index);
   index += 2;

   int maxLocals = ValueDecoder.readU2(buffer, index);
   index += 2;

   // NOTE: the opcode length is an unsigned 32 bit value in the specs.

   int opCodeLength = ValueDecoder.readU4(buffer, index);
   if(opCodeLength < 0) throw new ClassException(error + index + " [op-code length < 0]");
   index += 4;

   index += opCodeLength;
   int exceptionElements = ValueDecoder.readU2(buffer, index);
   index += 2;

   if(exceptionElements > 0) {
    ExceptionElement[] exceptionTable = new ExceptionElement[exceptionElements];

    for(int i = 0; i < exceptionElements; i++) {
     exceptionTable[i] = new ExceptionElement(buffer, index);
     index += ExceptionElement.size;
     }
    }

   int attributes = ValueDecoder.readU2(buffer, index);
   index += 2;

   if(attributes > 0) {
    Attribute[] table = new Attribute[attributes];

    for(int i = 0; i < attributes; i++) {
     AttributeFactory factory = new AttributeFactory(buffer, index);
     table[i] = factory.record;
     index = factory.offset;
     }
    }

   // check that the size reported matches the size read

   if(size != (index - orgIndex)) {
    throw new ClassException(error + index + " [size mismatch]");
    }

   // update the size and type

   size += 4;
   type = Attribute.CODE;
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException(error + index);
   }
  }

 /**
  * Gets the attribute as a string.
 **/

 public String stringize() {
  return "Code";
  }
 }
