//
// ConstantFieldRef.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

/**
 * A container for a CONSTANT_Fieldref record in a Java <tt>.class</tt> file.
 *
 * @version 1.1, 01 September 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John Leach</a>
 * @see ConstantFactory
 * @see ConstantUtf8
 * @see ConstantUnicode
 * @see ConstantInteger
 * @see ConstantFloat
 * @see ConstantLong
 * @see ConstantDouble
 * @see ConstantClass
 * @see ConstantString
 * @see ConstantMethodRef
 * @see ConstantInterfaceMethodRef
 * @see ConstantNameAndType
**/

public final class ConstantFieldRef extends Constant {

 /** The class index. */
 int classIndex;

 /** The name and type index. */
 int nameAndType;

 /** The exception error message. */
 private static final String error = "Fieldref record corrupt: ";

 /** The field name as a string. */
 private String name;

 /** The field reference as a string. */
 private String ref;

 /** The field reference signature. */
 private Signature sig;

 /**
  * Creates a CONSTANT_Fieldref record from a byte array.
  * @param buffer     the byte array
  * @param index      the offset in the input buffer
  * @exception ClassException	    if the record cannot be decoded
 **/

 public ConstantFieldRef(byte[] buffer, int index) throws ClassException {
  try {
   classIndex = ValueDecoder.readU2(buffer, index);
   index += 2;

   nameAndType = ValueDecoder.readU2(buffer, index);

   // update the size and type

   size = 4 + 1;
   type = Constant.FIELD_REF;
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException(error + index);
   }
  }

 /**
  * Gets the type signature.
  * @exception ClassException	    if the record cannot be decoded
 **/

 public Signature getSignature() throws ClassException {
  if(sig == null) {
   Constant type;

   try {
    type = constantPool[nameAndType];
    }
   catch(NullPointerException e) {
    throw new ClassException(error + "[no entry for name]");
    }
   catch(ArrayIndexOutOfBoundsException e) {
    throw new ClassException(error + "[no entry for name]");
    }

   if(type.type == Constant.NAME_AND_TYPE) sig = ((ConstantNameAndType)type).getSignature();
   else throw new ClassException(error + "[wrong type for name]");
   }
  return sig;
  }

 /**
  * Gets the CONSTANT_Fieldref class and name as a string.
  * @exception ClassException	    if the record cannot be decoded
 **/

 public String getName() throws ClassException {
  if(name == null) {
   String className, fieldName;
   Constant type;

   try {
    type = constantPool[classIndex];
    }
   catch(NullPointerException e) {
    throw new ClassException(error + "[no entry for class]");
    }
   catch(ArrayIndexOutOfBoundsException e) {
    throw new ClassException(error + "[no entry for class]");
    }

   if(type.type == Constant.CLASS) className = type.stringize();
   else throw new ClassException(error + "[wrong type for class]");

   try {
    type = constantPool[nameAndType];
    }
   catch(NullPointerException e) {
    throw new ClassException(error + "[no entry for name]");
    }
   catch(ArrayIndexOutOfBoundsException e) {
    throw new ClassException(error + "[no entry for name]");
    }

   if(type.type == Constant.NAME_AND_TYPE) fieldName = ((ConstantNameAndType)type).getName();
   else throw new ClassException(error + "[wrong type for name]");

   name = className + "." + fieldName;
   }

  return name;
  }

 /**
  * Gets the CONSTANT_Fieldref as a string.
  * @exception ClassException	    if the record cannot be decoded
 **/

 public String stringize() throws ClassException {
  if(ref == null) {
   if(name == null) getName();
   if(sig == null) getSignature();

   ref = sig.typeToString() + " " + name;
   }

  return ref;
  }
 }
