//
// ConstantInterfaceMethodRef.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

/**
 * A container for a CONSTANT_InterfaceMethodref record in a Java <tt>.class</tt>
 * file.
 *
 * @version 1.1, 01 September 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John leach</a>
 * @see ConstantFactory
 * @see ConstantUtf8
 * @see ConstantUnicode
 * @see ConstantInteger
 * @see ConstantFloat
 * @see ConstantLong
 * @see ConstantDouble
 * @see ConstantClass
 * @see ConstantString
 * @see ConstantFieldRef
 * @see ConstantMethodRef
 * @see ConstantNameAndType
**/

public final class ConstantInterfaceMethodRef extends Constant {

 /** The class index. */
 int classIndex;

 /** The name and type index. */
 int nameAndType;

 /** The exception error message. */
 private static final String error = "InterfaceMethodref record corrupt: ";

 /** The interface method name as a string. */
 private String name;

 /** The interface method reference as a string. */
 private String ref;

 /** The interface method reference signature. */
 private Signature sig;

 /**
  * Creates a CONSTANT_InterfaceMethodref record from a byte array.
  * @param buffer     the byte array
  * @param index      the offset in the input buffer
  * @exception ClassException	    if the record cannot be decoded
 **/

 public ConstantInterfaceMethodRef(byte[] buffer, int index) throws ClassException {
  try {
   classIndex = ValueDecoder.readU2(buffer, index);
   index += 2;

   nameAndType = ValueDecoder.readU2(buffer, index);

   // update the size and type

   size = 4 + 1;
   type = Constant.INTERFACE_METHOD_REF;
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException(error + index);
   }
  }

 /**
  * Gets the type signature.
  * @exception ClassException	    if the record cannot be decoded
 **/

 public Signature getSignature() throws ClassException {
  if(sig == null) {
   Constant type;

   try {
    type = constantPool[nameAndType];
    }
   catch(NullPointerException e) {
    throw new ClassException(error + "[no entry for name]");
    }
   catch(ArrayIndexOutOfBoundsException e) {
    throw new ClassException(error + "[no entry for name]");
    }

   if(type.type == Constant.NAME_AND_TYPE) sig = ((ConstantNameAndType)type).getSignature();
   else throw new ClassException(error + "[wrong type for name]");
   }
  return sig;
  }

 /**
  * Gets the CONSTANT_InterfaceMethodref class and name as a string.
  * @exception ClassException	    if the record cannot be decoded
 **/

 public String getName() throws ClassException {
  if(name == null) {
   String className, methodName;
   Constant type;

   try {
    type = constantPool[classIndex];
    }
   catch(NullPointerException e) {
    throw new ClassException(error + "[no entry for class]");
    }
   catch(ArrayIndexOutOfBoundsException e) {
    throw new ClassException(error + "[no entry for class]");
    }

   if(type.type == Constant.CLASS) className = type.stringize();
   else throw new ClassException(error + "[wrong type for class]");

   try {
    type = constantPool[nameAndType];
    }
   catch(NullPointerException e) {
    throw new ClassException(error + "[no entry for name]");
    }
   catch(ArrayIndexOutOfBoundsException e) {
    throw new ClassException(error + "[no entry for name]");
    }

   if(type.type == Constant.NAME_AND_TYPE) methodName = ((ConstantNameAndType)type).getName();
   else throw new ClassException(error + "[wrong type for name]");

   name = className + "." + methodName;
   }

  return name;
  }

 /**
  * Gets the CONSTANT_InterfaceMethdref as a string.
  * @exception ClassException	    if the record cannot be decoded
 **/

 public String stringize() throws ClassException {
  if(ref == null) {
   if(name == null) getName();
   if(sig == null) getSignature();

   // constructors are 'void' but we don't need to say so

   if(name.endsWith(".<init>")) ref = name + sig.paramsToString();
   else ref = sig.typeToString() + " " + name + sig.paramsToString();
   }

  return ref;
  }
 }
