//
// ConstantValueAttribute.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

/**
 * Decodes a ConstantValue attribute record in a Java <tt>.class</tt> file
 *
 * @version 1.0, 02 September 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John Leach</a>
 * @see Attribute
 * @see AttributeFactory
 * @see SourceFileAttribute
 * @see CodeAttribute
 * @see ExceptionsAttribute
 * @see LineNumberTableAttribute
 * @see LocalVariableTableAttribute
 * @see UnknownAttribute
**/

public final class ConstantValueAttribute extends Attribute {

 /** The constant value index. */
 int valueIndex;

 /** The constant value type. */
 Constant constantType;

 /** The exception error message. */
 private static final String error = "ConstantValue attribute record corrupt: ";

 /** The attribute value. */
 private String attributeValue;

 /**
  * Creates a ConstantValue attribute record from a byte array.
  * @param buffer     the byte array
  * @param index      the offset in the input buffer
  * @exception ClassException	if the record cannot be decoded
 **/

 public ConstantValueAttribute(byte[] buffer, int index) throws ClassException {
  try {

   // NOTE: the attribute size is an unsigned 32 bit value in the specs.

   int attrSize = ValueDecoder.readU4(buffer, index);
   if(attrSize != 2) throw new ClassException(error + index + " [size != 2]");
   index += 4;

   valueIndex = ValueDecoder.readU2(buffer, index);
   try {
    constantType = Constant.constantPool[valueIndex];

    if((constantType.type != Constant.INTEGER) &&
       (constantType.type != Constant.LONG) &&
       (constantType.type != Constant.STRING) &&
       (constantType.type != Constant.FLOAT) &&
       (constantType.type != Constant.DOUBLE)) throw new ClassException(error + index + " [wrong type]");
    }
   catch(NullPointerException e) {
    throw new ClassException(error + index + " [no entry]");
    }
   catch(ArrayIndexOutOfBoundsException e) {
    throw new ClassException(error + index + " [no entry]");
    }

   // update the size and type

   size = 4 + 2;
   type = Attribute.CONSTANT_VALUE;
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException(error + index);
   }
  }

 /**
  * Gets the attribute as a string.
  * @exception ClassException	    if the record cannot be decoded
 **/

 public String stringize() throws ClassException {
  if(attributeValue == null) attributeValue = constantType.stringize();
  return attributeValue;
  }
 }
