//
// ExceptionsAttribute.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

/**
 * Decodes an Exceptions attribute record in a Java <tt>.class</tt> file
 *
 * @version 1.0, 02 September 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John Leach</a>
 * @see Attribute
 * @see AttributeFactory
 * @see SourceFileAttribute
 * @see ConstantValueAttribute
 * @see CodeAttribute
 * @see LineNumberTableAttribute
 * @see LocalVariableTableAttribute
 * @see UnknownAttribute
**/

public final class ExceptionsAttribute extends Attribute {

 /** The Exceptions count. */
 int exceptionCount;

 /** The Exception Pool. */
 int exceptionPool[];

 /** The attribute value. */
 private String attributeValue;

 /** The exception error message. */
 private static final String error = "Exceptions attribute record corrupt: ";

 /**
  * Creates an Exceptions attribute record from a byte array.
  * @param buffer     the byte array
  * @param index      the offset in the input buffer
  * @exception ClassException	if the record cannot be decoded
 **/

 public ExceptionsAttribute(byte[] buffer, int index) throws ClassException {
  try {

   // NOTE: size is a long in the specs.

   size = (int)ValueDecoder.readU4(buffer, index);
   index += 4;

   exceptionCount = ValueDecoder.readU2(buffer, index);
   index += 2;

   if(exceptionCount > 0) {
    exceptionPool = new int[exceptionCount];

    for(int i = 0; i < exceptionCount; i++) {
     int exceptionIndex;
     exceptionIndex = ValueDecoder.readU2(buffer, index);
     exceptionPool[i] = exceptionIndex;

     try {
      if(exceptionIndex != 0 && Constant.constantPool[exceptionIndex].type != Constant.CLASS) throw new ClassException(error + index + " [wrong type]");
      }
     catch(NullPointerException e) {
      throw new ClassException(error + index + " [no entry]");
      }
     catch(ArrayIndexOutOfBoundsException e) {
      throw new ClassException(error + index + " [no entry]");
      }

     index += 2;
     }
    }

   // update the size and type

   size += 4;
   type = Attribute.EXCEPTIONS;
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException(error + index);
   }
  }

 /**
  * Gets the attribute as a string.
 **/

 public String stringize() throws ClassException {
  if(attributeValue == null) {
   for(int i = 0; i < exceptionCount; i++) {
    int exceptionIndex = exceptionPool[i];
    String exceptionName = "java.lang.Exception";
    if(exceptionIndex != 0) exceptionName = Constant.constantPool[exceptionIndex].stringize();

    if(attributeValue == null) attributeValue = exceptionName;
    else attributeValue = attributeValue + ", " + exceptionName;
    }
   }

  return attributeValue;
  }
 }
