//
// Field.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

/**
 * A container for a field record in a Java <tt>.class</tt> file.
 *
 * @version 1.1, 02 September 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John Leach</a>
 * @see Attribute
 * @see Method
**/

public final class Field {

 /** The field access flags. */
 Access access;

 /** The constant pool index to the field name. */
 int nameIndex;

 /** The constant pool index to the signature name. */
 int signatureIndex;

 /** The number of attributes of this field. */
 int attributeCount;

 /** The attribute table of this field. */
 Attribute table[];

 /** The field field length. */
 int size;

 /** The exception error message. */
 private static final String error = "Field record corrupt: ";

 /** The complete field specification as a string. */
 private String fieldSpec;

 /** The field signature. */
 private Signature sig;

 /**
  * Creates a field record from a byte array.
  * @param buffer     the byte array
  * @param index      the offset in the input buffer
  * @exception ClassException	if the record cannot be decoded
 **/

 public Field(byte[] buffer, int index) throws ClassException {
  access = new Access(buffer, index);
  index += access.size;
  size = access.size;

  try {
   Constant type;
   nameIndex = ValueDecoder.readU2(buffer, index);
   try {
    type = Constant.constantPool[nameIndex];
    }
   catch(NullPointerException e) {
    throw new ClassException(error + index + " [no entry for name]");
    }
   catch(ArrayIndexOutOfBoundsException e) {
    throw new ClassException(error + index + " [no entry for name]");
    }
   if(type.type != Constant.UTF8) throw new ClassException(error + index + " [wrong type for name]");

   index += 2;
   size += 2;

   signatureIndex = ValueDecoder.readU2(buffer, index);
   try {
    type = Constant.constantPool[signatureIndex];
    }
   catch(NullPointerException e) {
    throw new ClassException(error + index + " [no entry for signature]");
    }
   catch(ArrayIndexOutOfBoundsException e) {
    throw new ClassException(error + index + " [no entry for signature]");
    }
   if(type.type != Constant.UTF8) throw new ClassException(error + index + " [wrong type for signature]");

   index += 2;
   size += 2;

   attributeCount = ValueDecoder.readU2(buffer, index);
   index += 2;
   size += 2;
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException(error + index);
   }

  if(attributeCount > 0) {
   table = new Attribute[attributeCount];
   for(int i = 0; i < attributeCount; i++) {
    AttributeFactory factory = new AttributeFactory(buffer, index);
    table[i] = factory.record;
    index += factory.size;
    size += factory.size;
    }
   }
  }

 /**
  * Gets the field signature.
  * @exception ClassException	    if the record cannot be decoded
 **/

 public Signature getSignature() throws ClassException {
  if(sig == null) {
   Constant type = Constant.constantPool[signatureIndex];
   sig = new Signature(type.stringize());
   }
  return sig;
  }

 /**
  * Gets the field as a string.
  * @exception ClassException	    if the record cannot be decoded
 **/

 public String stringize() throws ClassException {
  if(fieldSpec == null) {
   fieldSpec = access.stringize();

   Constant type = Constant.constantPool[nameIndex];
   String name = type.stringize();

   type = Constant.constantPool[signatureIndex];
   sig = new Signature(type.stringize());

   fieldSpec = fieldSpec + sig.typeToString() + " " + name;
   fieldSpec = fieldSpec + getAttributes();
   }
  return fieldSpec;
  }

 /**
  * Gets the attributes for the field.
  * @return	the attribute values as a string
  * @exception ClassException	    if the record cannot be decoded
 **/

 private String getAttributes() throws ClassException {
  String text = "";
  for(int i = 0; i < attributeCount; i++) {
   Attribute attribute = table[i];
   if(attribute != null && text.equals("") && attribute.type == Attribute.CONSTANT_VALUE) {
    text = " = " + attribute.stringize();
    }
   }
  return text;
  }
 }
