//
// LineNumberTableAttribute.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

/**
 * Decodes a LineNumberTable attribute record in a Java <tt>.class</tt> file
 *
 * @version 1.0, 02 September 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John Leach</a>
 * @see LineNumberElement
 * @see Attribute
 * @see AttributeFactory
 * @see SourceFileAttribute
 * @see ConstantValueAttribute
 * @see CodeAttribute
 * @see ExceptionsAttribute
 * @see LocalVariableTableAttribute
 * @see UnknownAttribute
**/

public final class LineNumberTableAttribute extends Attribute {

 /** The number of elements. */
 int elements;

 /** The line number table. */
 LineNumberElement[] table;

 /** The exception error message. */
 private static final String error = "LineNumberTable attribute record corrupt: ";

 /**
  * Creates a LineNumberTable attribute record from a byte array.
  * @param buffer     the byte array
  * @param index      the offset in the input buffer
  * @exception ClassException	if the record cannot be decoded
 **/

 public LineNumberTableAttribute(byte[] buffer, int index) throws ClassException {
  try {

   // NOTE: the size is an unsigned 32 bit value in the specs.

   size = ValueDecoder.readU4(buffer, index);
   index += 4;

   // used to check the size declared

   int orgIndex = index;

   elements = ValueDecoder.readU2(buffer, index);
   index += 2;

   if(elements > 0) {
    table = new LineNumberElement[elements];

    for(int i = 0; i < elements; i++) {
     table[i] = new LineNumberElement(buffer, index);
     index += LineNumberElement.size;
     }
    }

   // check the declared size against the size read

   if(size != (index - orgIndex)) {
    throw new ClassException(error + index + " [size mismatch]");
    }

   // update the size and type

   size += 4;
   type = Attribute.LINE_NUMBER_TABLE;
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException(error + index);
   }
  }

 /**
  * Gets the attribute as a string.
 **/

 public String stringize() {
  return "LineNumberTable";
  }
 }
