//
// LocalVariableTableAttribute.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

import java.util.Vector;

/**
 * Decodes a LocalVariableTable attribute record in a Java <tt>.class</tt> file
 *
 * @version 1.0, 02 September 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John Leach</a>
 * @see LocalVariableElement
 * @see Attribute
 * @see AttributeFactory
 * @see SourceFileAttribute
 * @see ConstantValueAttribute
 * @see CodeAttribute
 * @see ExceptionsAttribute
 * @see LineNumberTableAttribute
 * @see UnknownAttribute
**/

public final class LocalVariableTableAttribute extends Attribute {

 /** The number of elements. */
 int elements;

 /** The local variable table. */
 Vector table;

 /** The exception error message. */
 private static final String error = "LocalVariableTable attribute record corrupt: ";

 /**
  * Creates a LocalVariableTable attribute record from a byte array.
  * @param buffer     the byte array
  * @param index      the offset in the input buffer
  * @exception ClassException	if the record cannot be decoded
 **/

 public LocalVariableTableAttribute(byte[] buffer, int index) throws ClassException {
  try {

   // NOTE: the size is an unsigned 32 bit value in the specs.

   size = ValueDecoder.readU4(buffer, index);
   index += 4;

   // used to check the declared size

   int orgIndex = index;

   int count = ValueDecoder.readU2(buffer, index);
   index += 2;

   if(count > 0) {
    elements = count;
    table = new Vector(elements);

    LocalVariableElement elem;
    for(int i = 0; i < elements; i++) {
     elem = new LocalVariableElement(buffer, index);
     addLocalVariable(elem);
     index += elem.size;
     }
    }

   // check the declared size against the size read

   if(size != (index - orgIndex)) {
    throw new ClassException(error + index + " [size mismatch]");
    }

   // update the size and type

   size += 4;
   type = Attribute.LOCAL_VARIABLE_TABLE;
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException(error + index);
   }
  }

 /**
  * Adds a local variable reference to the table.
  * @param ref	the local variable to add.
 **/

 public void addLocalVariable(LocalVariableElement ref) {
  int slotIndex;
  int pcIndex;
  LocalVariableElement elem;

  // order by slot ...

  for(slotIndex = 0; slotIndex < table.size(); slotIndex++) {
   elem = (LocalVariableElement)table.elementAt(slotIndex);

   if(elem.slotNumber > ref.slotNumber) {
    table.insertElementAt(ref, slotIndex);
    return;
    }
   else if(elem.slotNumber == ref.slotNumber) {

    // ... then program counter

    for(pcIndex = slotIndex; pcIndex < table.size(); pcIndex++) {
     elem = (LocalVariableElement)table.elementAt(pcIndex);

     if(elem.slotNumber == ref.slotNumber) {
      if(elem.startPC > ref.startPC) {

       table.insertElementAt(ref, pcIndex);
       return;
       }
      }
     else {

      table.insertElementAt(ref, pcIndex);
      return;
      }
     }

    // if we get this far, then add to the end of the slot

    table.addElement(ref);
    return;
    }
   }

  // if we get this far, then add to the end of the table

  table.addElement(ref);
  }

 /**
  * Gets the attribute as a string.
 **/

 public String stringize() {
  return "LocalVariableTable";
  }
 }
