//
// Parser.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

import java.io.*;

/**
 * This class reads a Java <tt>.class</tt> file via an input stream and
 * parses it into textual information, which is then stored in an
 * instance of ClassData.
 *
 * @version 1.1, 10 October 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John Leach</a>
 * @see ClassData
**/

public class Parser {

 /** Contains the parsed class information. */
 protected ClassData classData;

 /** The class file image in memory. */
 protected byte[] input;

 /** The constant buffer expansion size. */
 private final int bufferSize = 1024 * 4;

 /** The compiler minor version number. */
 protected int minor;

 /** The compiler major version number. */
 protected int major;

 /** The class name index. */
 protected int classIndex;

 /** The class name. */
 protected String className;

 /** The class package. */
 protected String packageName;

 /** The super class name. */
 protected String superName;

 /** The super class name index. */
 protected int superIndex;

 /** The list of other (imported) classes used. */
 protected String usedClasses[];

 /** The list of fields referenced. */
 protected String fieldRefs[];

 /** The list of methods referenced. */
 protected String methodRefs[];

 /** The list of interface methods referenced. */
 protected String interfaceMethodRefs[];

 /** The list of class fields. */
 protected String fieldNames[];

 /** The list of class methods. */
 protected String methodNames[];

 /** The source file name. */
 protected String sourceFilename;

 /** The full class specification. */
 protected String classSpec;

 /**
  * Reads an input stream and decodes the class information from that
  * stream into an instance of ClassData.
  * @param dataIn	    the input class stream
  * @param classData  the parsed output class data
  * @exception ClassException	if the parser fails to decode the stream
 **/

 public Parser(InputStream dataIn, ClassData classData) throws ClassException {
  this.classData = classData;
  try {
   readInput(dataIn);
   }
  catch(IOException e) {
   throw new ClassException(e.getMessage());
   }
  decode();
  setClassData();
  }

 /**
  * Decodes the class information from a byte array into an instance of
  * ClassData.
  * @param buffer	    the input byte array
  * @param classData  the parsed output class data
  * @exception ClassException	if the parser fails to decode the stream
 **/

 public Parser(byte[] buffer, ClassData classData) throws ClassException {
  this.classData = classData;
  input = buffer;
  decode();
  setClassData();
  }

 /**
  * Allows sub-classing of the constructor.
 **/

 protected Parser() { }

 /**
  * Reads the class from the input stream into the bytecode buffer.
  * @param dataIn	the InputStream to be read
  * @exception IOException   if the parser fails to read the stream
 **/

 protected final void readInput(InputStream dataIn) throws IOException {
  input = new byte[bufferSize];

  int inputUsed = 0, n;
  while((n = dataIn.read(input, inputUsed, input.length - inputUsed)) >= 0) {
   inputUsed += n;
   if(inputUsed == input.length) {
    byte newinput[] = new byte[inputUsed + bufferSize];
    System.arraycopy(input, 0, newinput, 0, inputUsed);
    input = newinput;
    }
   }
  byte newinput[] = new byte[inputUsed];
  System.arraycopy(input, 0, newinput, 0, inputUsed);
  input = newinput;
  }

 /**
  * Decodes the class in the input buffer into textual class information.
  * @exception ClassException	if the parser fails to decode the stream
 **/

 protected void decode() throws ClassException {

  // first check that this is a .class file CAFE BABE - very original!

  try {
   int magic = ValueDecoder.readU4(input, 0);
   if(magic != 0xCAFEBABE) throw new ClassException("Not a .class file");
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException("Not a .class file");
   }

  // compiler version

  try {
   minor = ValueDecoder.readU2(input, 4);
   major = ValueDecoder.readU2(input, 6);
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException("Version number corrupt: 4");
   }

  // constant pool

  int offset = 10;
  int constantPoolCount;
  Constant constantPool[] = null;
  try {
   constantPoolCount = ValueDecoder.readU2(input, 8);
   constantPool = new Constant[constantPoolCount];
   ConstantFactory factory = new ConstantFactory(constantPool, input, offset);
   offset = factory.offset;
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException("Constant pool corrupt: " + offset);
   }

  // access flags

  Access access = new Access(input, offset);
  offset += access.size;

  // class name

  try {
   classIndex = ValueDecoder.readU2(input, offset);
   if(constantPool[classIndex] == null || constantPool[classIndex].type != Constant.CLASS)
    throw new ClassException("thisClass record corrupt: " + offset);
   className = constantPool[classIndex].stringize();
   offset += 2;
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException("thisClass record corrupt: " + offset);
   }

  // get the package name

  int charIndex = className.lastIndexOf('/');
  if(charIndex >= 1) packageName = className.substring(0, charIndex + 1);
  else packageName = null;

  // super class name

  superName = "java/lang/Object";
  try {
   superIndex = ValueDecoder.readU2(input, offset);
   if(superIndex != 0 && (constantPool[superIndex] == null || constantPool[superIndex].type != Constant.CLASS))
    throw new ClassException("superClass record corrupt: " + offset);

   if(superIndex != 0) superName = constantPool[superIndex].stringize();
   offset += 2;
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException("superClass record corrupt: " + offset);
   }

  // classes, fields, methods and interface methods referenced

  int classCount = 0;
  int fieldRefCount = 0;
  int methodRefCount = 0;
  int interfaceMethodRefCount = 0;
  for(int i = 1; i < constantPoolCount; i++) {
   Constant type = constantPool[i];
   if(type != null) {
    if(type.type == Constant.CLASS && i != classIndex) ++classCount;
    if(type.type == Constant.FIELD_REF) ++fieldRefCount;
    else if(type.type == Constant.METHOD_REF) ++methodRefCount;
    else if(type.type == Constant.INTERFACE_METHOD_REF) ++interfaceMethodRefCount;
    }
   }

  if(classCount > 0) usedClasses = new String[classCount];
  if(fieldRefCount > 0) fieldRefs = new String[fieldRefCount];
  if(methodRefCount > 0) methodRefs = new String[methodRefCount];
  if(interfaceMethodRefCount > 0) interfaceMethodRefs = new String[interfaceMethodRefCount];
  classCount = 0;
  fieldRefCount = 0;
  methodRefCount = 0;
  interfaceMethodRefCount = 0;

  for(int i = 1; i < constantPoolCount; i++) {
   Constant type = constantPool[i];
   if(type != null) {
    if(type.type == Constant.CLASS && i != classIndex) {
     usedClasses[classCount] = type.stringize();
     ++classCount;
     }
    if(type.type == Constant.FIELD_REF) {
     fieldRefs[fieldRefCount] = type.stringize();
     ++fieldRefCount;
     }
    else if(type.type == Constant.METHOD_REF) {
     methodRefs[methodRefCount] = checkConstructor(type.stringize());
     ++methodRefCount;
     }
    else if(type.type == Constant.INTERFACE_METHOD_REF) {
     interfaceMethodRefs[interfaceMethodRefCount] = checkConstructor(type.stringize());
     ++interfaceMethodRefCount;
     }
    }
   }

  // interfaces

  String interfaceSpec = null;
  try {
   int interfaceCount = ValueDecoder.readU2(input, offset);
   offset += 2;
   if(interfaceCount > 0) {
    for(int i = 0; i < interfaceCount; i++) {
     int index = ValueDecoder.readU2(input, offset);
     Constant type = constantPool[index];
     if(type != null && type.type == Constant.CLASS) {
      if(interfaceSpec == null) interfaceSpec = " implements " + type.stringize();
      else interfaceSpec = interfaceSpec + ", " + type.stringize();
      }
     offset += 2;
     }
    }
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException("Interface record corrupt: " + offset);
   }
  if(interfaceSpec == null) interfaceSpec = "";

  // fields

  try {
   int fieldCount = ValueDecoder.readU2(input, offset);
   offset += 2;
   if(fieldCount > 0) {
    fieldNames = new String[fieldCount];
    for(int i = 0; i < fieldCount; i++) {
     Field field = new Field(input, offset);
     fieldNames[i] = field.stringize();
     offset += field.size;
     }
    }
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException("Field record corrupt: " + offset);
   }

  // methods

  try {
   int methodCount = ValueDecoder.readU2(input, offset);
   offset += 2;
   if(methodCount > 0) {
    methodNames = new String[methodCount];
    for(int i = 0; i < methodCount; i++) {
     Method method = new Method(input, offset);
     methodNames[i] = checkConstructor(method.stringize());
     offset += method.size;
     }
    }
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException("Method record corrupt: " + offset);
   }

  // attributes

  sourceFilename = "";
  try {
   int attributeCount = ValueDecoder.readU2(input, offset);
   offset += 2;
   if(attributeCount > 0) {
    for(int i = 0; i < attributeCount; i++) {
     AttributeFactory factory = new AttributeFactory(input, offset);
     if(factory.record.type == Attribute.SOURCE_FILE) sourceFilename = factory.record.stringize();
     offset += factory.size;
     }
    }
   }
  catch(ArrayIndexOutOfBoundsException e) {
   throw new ClassException("Attribute record corrupt: " + offset);
   }

  // build the class specifications

  classSpec = access.stringize();
  if(access.isInterface() == false) classSpec = classSpec + "class ";
  classSpec = classSpec + className;
  if(superIndex > 0) classSpec = classSpec + " extends " + superName;
  classSpec = classSpec + interfaceSpec;
  }

 /**
  * Sets the ClassData object with the information parsed
 **/

 protected void setClassData() {
  classData.setVersion(Integer.toString(major) + "." + Integer.toString(minor));
  classData.setClassName(className);
  classData.setSuperClassName(superName);
  classData.setUsedClasses(usedClasses);
  classData.setFieldRefs(fieldRefs);
  classData.setMethodRefs(methodRefs);
  classData.setInterfaceMethodRefs(interfaceMethodRefs);
  classData.setClassFields(fieldNames);
  classData.setClassMethods(methodNames);
  classData.setSourceFileName(sourceFilename);
  classData.setClassSpec(classSpec);
  }

 /**
  * Substitutes the '.&lt;init&gt;' string with nothing (constructor).
  * <br>
  * Substitutes the '&lt;init&gt;' string with the class name (constructor).
  * <br>
  * Substitutes the '&lt;clinit&gt;' string with null (static class constructor).
  * @param str	the input string
 **/

 protected String checkConstructor(String str) {

  // general constructors

  int index = str.indexOf(".<init>");
  if(index >= 0) return str.substring(0, index) + str.substring(index + 7);

  // class constructors

  index = str.indexOf("<init>");
  if(index >= 0) return str.substring(0, index) + className + str.substring(index + 6);

  // static class initialisers can be ignored

  index = str.indexOf("<clinit>");
  if(index >= 0) return null;

  return str;
  }
 }
