//
// Signature.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

import java.util.Vector;

/**
 * A container and parser for a signature record (CONSTANT_Utf8 record) in a 
 * <tt>.class</tt> file.
 *
 * @version 1.1, 03 September 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John Leach</a>
**/

public final class Signature {

 /** The original signature. */
 String sig;

 /** The parsed type. */
 String type;

 /** The number of parameters in the signature. */
 int parameters;

 /** The parameter types as a Vector of Strings. */
 Vector params;

 /** A flag to indicate if the return type is void. */
 boolean voidReturn;

 /**
  * Parses a signature from a signature string.
  * @param sig	      the signature string
  * @exception ClassException	    if the record cannot be decoded
 **/

 public Signature(String sig) throws ClassException {
  this.sig = sig;

  parameters = 0;
  voidReturn = false;
  int index = 0;
  int length = 0;

  // parse the parameters

  if(index < sig.length() && sig.charAt(index) == '(') {
   ++index;
   while(index < sig.length() && sig.charAt(index) != ')') {
    length = parse(sig, index);
    if(params == null) params = new Vector();
    params.addElement(sig.substring(index, index + length));
    ++parameters;
    index += length;
    }
   if(index < sig.length() && sig.charAt(index) == ')') ++index;
   }

  // parse the return type

  length = parse(sig, index);
  type = sig.substring(index, index + length);
  index += length;

  if(index < sig.length()) throw new ClassException("Internal error: Extra text found in signature '" + sig + "'");

  if(type.charAt(0) == 'V') voidReturn = true;
  }

 /**
  * Converts the type signature to a String.
  * @exception ClassException	    if the record cannot be decoded
 **/

 public String typeToString() throws ClassException {
  if(type == null) throw new ClassException("Internal error: Missing signature");
  return typeToString(type);
  }

 /**
  * Converts the parameter type signatures to a String.
  * @exception ClassException	    if the record cannot be decoded
 **/

 public String paramsToString() throws ClassException {
  String result = "(";
  String type;

  for(int i = 0; i < (parameters - 1); i++) {
   type = (String)params.elementAt(i);
   result = result + typeToString(type) + ", ";
   }
  if(parameters > 0) {
   type = (String)params.elementAt(parameters - 1);
   result = result + typeToString(type);
   }

  return result + ")";
  }

 /**
  * Parses the signature into types.
  * @param sig	      the signature string
  * @param index      the current index
  * @return	      the type signature length
  * @exception ClassException	    if the record cannot be decoded
 **/

 private int parse(String sig, int index) throws ClassException {
  int length = 0;

  while(index < sig.length()) {
   int ch = sig.charAt(index);
   ++index;
   ++length;

   switch(ch) {
    case 'B':
    case 'C':
    case 'D':
    case 'F':
    case 'I':
    case 'J':
    case 'S':
    case 'V':
    case 'Z': return length;

    case 'L':
    while(index < sig.length() && sig.charAt(index) != ';') {
     ++index;
     ++length;
     }
    if(index < sig.length() && sig.charAt(index) == ';') {
     ++index;
     ++length;
     }
    return length;

    case '[':
    while(index < sig.length() && (sig.charAt(index) >= '0' && sig.charAt(index) <= '9')) {
     ++index;
     ++length;
     }
    break;

    default: throw new ClassException("Internal error: Unknown signature '" + (char)ch + "'");
    }
   }

  return length;
  }

 /**
  * Converts a type signature to a String.
  * @param sig	      the type signature string
  * @exception ClassException	    if the record cannot be decoded
 **/

 private static String typeToString(String sig) throws ClassException {
  int index = 0;
  int start;
  String ret = "";
  String array = "";

  while(index < sig.length()) {
   int ch = sig.charAt(index);
   ++index;

   switch(ch) {
    case 'B': ret = "byte";    return ret + array;
    case 'C': ret = "char";    return ret + array;
    case 'D': ret = "double";  return ret + array;
    case 'F': ret = "float";   return ret + array;
    case 'I': ret = "int";     return ret + array;
    case 'J': ret = "long";    return ret + array;
    case 'S': ret = "short";   return ret + array;
    case 'V': ret = "void";    return ret + array;
    case 'Z': ret = "boolean"; return ret + array;

    case 'L':
    start = index;
    while(index < sig.length() && sig.charAt(index) != ';') ++index;
    ret = sig.substring(start, index);
    return ret + array;

    case '[':
    start = index;
    while(index < sig.length() && (sig.charAt(index) >= '0' && sig.charAt(index) <= '9')) ++index;
    array = array + "[";
    if(index > start) array = array + sig.substring(start, index);
    array = array + "]";
    break;

    default: throw new ClassException("Internal error: Unknown signature '" + (char)ch + "'");
    }
   }

  return ret + array;
  }
 }
