//
// ValueDecoder.java
//
// Copyright (c) 1996 John Leach. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for NON-COMMERCIAL purposes and without fee is hereby granted
// provided that this copyright notice appears in all copies.
//
// JOHN LEACH MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
// THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
// NON-INFRINGEMENT. JOHN LEACH SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
// BY A LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
// OR ITS DERIVATIVES.
//

package jclass;

/**
 * A generic Java <tt>.class</tt> file value decoder.
 *
 * @version 1.0, 02 September 1996
 * @author  <a href="mailto:leachj@easy1.easynet.it">John Leach</a>
**/

public final class ValueDecoder {

 /**
  * This class cannot be instantiated.
 **/

 private ValueDecoder() { }

 /**
  * Reads a U2 (unsigned 16 bit) value.
  * @param buffer  the input buffer
  * @param index   the offset in the input buffer
  * @return	   the u2 value expressed as an int
  *
  * @exception ArrayIndexOutOfBoundsException if less than two bytes remain
  *					      in the array
 **/

 public static int readU2(byte[] buffer, int index) throws ArrayIndexOutOfBoundsException {
  return ((buffer[index] & 0xFF) << 8) + (buffer[index + 1] & 0xFF);
  }

 /**
  * Reads an S2 (signed 16 bit) value.
  * @param buffer  the input buffer
  * @param index   the offset in the input buffer
  * @return	   the s2 value expressed as an int
  *
  * @exception ArrayIndexOutOfBoundsException if less than two bytes remain
  *					      in the array
 **/

 public static int readS2(byte[] buffer, int index) throws ArrayIndexOutOfBoundsException {
  return ((int)buffer[index] << 8) + (buffer[index + 1] & 0xFF);
  }

 /**
  * Reads a U4 (unsigned 32 bit) value.
  * @param buffer  the input buffer
  * @param index   the offset in the input buffer
  * @return	   the u4 value expressed as an int
  *
  * @exception ArrayIndexOutOfBoundsException if less than four bytes remain
  *					      in the array
 **/

 public static int readU4(byte[] buffer, int index) throws ArrayIndexOutOfBoundsException {
  int high = readU2(buffer, index);
  int low = readU2(buffer, index + 2);
  return (high << 16) + low;
  }

 /**
  * Reads an S4 (signed 32 bit) value.
  * @param buffer  the input buffer
  * @param index   the offset in the input buffer
  * @return	   the s4 value expressed as an int
  *
  * @exception ArrayIndexOutOfBoundsException if less than four bytes remain
  *					      in the array
 **/

 public static int readS4(byte[] buffer, int index) throws ArrayIndexOutOfBoundsException {
  return readU4(buffer, index);
  }

 /**
  * Reads a U8 (unsigned 64 bit) value (not documented, but is the equivalent
  * of a 'long').
  * @param buffer  the input buffer
  * @param index   the offset in the input buffer
  * @return	   the u8 value expressed as a long
  *
  * @exception ArrayIndexOutOfBoundsException if less than eight bytes remain
  *					      in the array
 **/

 public static long readU8(byte[] buffer, int index) throws ArrayIndexOutOfBoundsException {
  long high = readU4(buffer, index);
  long low = readU4(buffer, index + 4);
  return (high << 32L) + low;
  }
 }
